require "entruby/version"
require "doxyparser"
require "find"

class Reference
  def initialize(element)
    @id = element[:refid]
    @content = element.content
  end

  # @return Doxyparser::Function
  def reveal(collection)
    collection.find { |member| member.basename == @content }
  end
end

module Doxyparser
  class Member
    def col
      aux = self.xpath("location")[0]
      bodyEnd, bodyStart = aux["bodyend"], aux["bodystart"]
      @col = bodyEnd && bodyStart ? Integer(bodyEnd) - Integer(bodyStart) : nil
    end
  end

  class Param
    def as_json(*)
      {type: {name: type.name, basename: type.basename}, name: declname}
    end
  end

  class Function
    # @return Array[Reference]
    def references
      node.css('references').map { |ref| Reference.new(ref) }
    end

    def referenced_by
      node.css('referencedby').map { |ref| Reference.new(ref) }
    end

    def efferent(all = nil)
      all.to_a.empty? ? references : references.map { |ref| ref.reveal(all) }.compact
    end

    def afferent(all = nil)
      all.to_a.empty? ? referenced_by : referenced_by.map { |ref| ref.reveal(all) }.compact
    end

    def fan_out(all = nil)
      efferent(all).size
    end

    def fan_in(all = nil)
      afferent(all).size
    end

    def coupling_metric(all = nil)
      fan_out = fan_out(all).to_f
      fan_in = fan_in(all).to_f
      denominator = fan_out + fan_in
      denominator == 0 ? 0 : fan_out / (denominator)
    end

    def global_name
      "global::#{signature}"
    end

    def valid_params
      params.reject { |p| p.type.name == 'void' }
    end

    def signature
      "#{basename}(#{valid_params.map { |p| p.type.name }.join(', ')})"
    end

    # @return Report
    def report(all)
      {id: id, name: basename, signature: signature, location: location, metric: coupling_metric(all), col: col,
       params: valid_params.map(&:as_json),
       imports: efferent(all).map(&:signature)}
    end
  end
end

module Entruby
  class Error < StandardError; end

  class OverviewMetric
    attr_reader :name, :description, :value, :weight

    def initialize(name, description, value, weight)
      @name, @description, @value, @weight = name, description, value, weight
    end

    def as_json(*args)
      {name: name, description: description, value: value, weight: weight}
    end
  end

  class Overview
    attr_reader :metrics

    COMPLEX_FUNCTION_PARAMS = 7
    HIGH_COUPLING_METRIC = 0.7
    HUGE_FUNCTION_COL = 50

    def initialize(couplings)
      huge_functions = couplings.count { |c| c[:col] >= HUGE_FUNCTION_COL }
      complex_functions = couplings.count { |c| c[:params].size >= COMPLEX_FUNCTION_PARAMS }
      couplers = couplings.count { |c| c[:metric] >= HIGH_COUPLING_METRIC }
      @metrics = [
          OverviewMetric.new("Excessive Couplers", "(Metric >= 0.7)", couplers, 3),
          OverviewMetric.new("Complex Functions", "(Params >= 7)", complex_functions, 2),
          OverviewMetric.new("Huge Functions", "(CoL >= 50)", huge_functions, 1),
      ]
    end

    def report
      {index: index, metrics: metrics.map(&:as_json)}
    end

    def index
      metrics.reduce(0.0) { |memo, metric| memo + metric.value * metric.weight }
    end
  end

  class App
    attr_reader :output_dir

    def initialize(argv)
      if argv.size < 2
        puts "USAGE: $ entruby <source directory> <doxygen xml directory>"
        exit 1
      end

      @source_dir, @xml_dir, @output_dir = argv[0], argv[1], Dir.pwd
    end

    # @return [Array<Doxyparser::Function>]
    def functions
      source_file_paths = []
      Find.find(@source_dir) do |path|
        source_file_paths << File.basename(path).gsub('_', '__') if path =~ /.*\.c$/
      end
      source_file_paths.reduce([]) do |functions, file|
        hfile = Doxyparser::parse_file(file, @xml_dir)
        functions + hfile.functions
      end
    end

    def functions_report
      all_functions = functions
      report = all_functions.map { |func| func.report(all_functions) }
      report.sort_by! { |row| row[:metric] }.reverse!
    end

    def overview(couplings)
      Overview.new(couplings).report
    end

    def report
      couplings = functions_report
      {overview: overview(couplings), functions: couplings}
    end
  end
end